# flake8: noqa
from aixd_ara.wrappers import WrapperShallowDataObject

if not dim or dim < 1:
    dim = 1

if domain:
    domain = (float(domain.Min), float(domain.Max))
else:
    domain = None


dobj = {"datatype": "DataReal", "name": name, "dim": dim, "domain": domain}


dataobject = WrapperShallowDataObject(dobj)



# flake8: noqa

# this code has been inspired by a forum post by "chanley" (2018.11.02)
# https://discourse.mcneel.com/t/can-i-instantiate-specific-component-on-the-canvas-with-a-script-python/74204/8

import Grasshopper
import System.Drawing as sd

from aixd_ara.gh_ui import get_dataobject_names_from_block
from aixd_ara.gh_ui_helper import session_id

comp = ghenv.Component
ghdoc = comp.OnPingDocument()


def make_Panel(NickName, UserText, Pivot, Bounds):
    try:
        Panel = Grasshopper.Kernel.Special.GH_Panel()
        Panel.NickName = NickName
        Panel.UserText = UserText
        Panel.Properties.Colour = sd.Color.White
        # Panel.Properties.Font = sd.Font("Trebuchet MS", 10)
        Panel.Properties.Multiline = False
        Panel.Properties.DrawIndices = False
        Panel.Properties.DrawPaths = False
        ghdoc.AddObject(Panel, False, ghdoc.ObjectCount + 1)
        Panel.Attributes.Pivot = Pivot
        Panel.Attributes.Bounds = Bounds
    except Exception(ex):
        ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, str(ex))


x = ghenv.Component.Attributes.DocObject.Attributes.Bounds.Right
y = ghenv.Component.Attributes.DocObject.Attributes.Bounds.Top

w = 150
h = 150
gap = 50

errors = ""

datablocks = ["design_parameters", "performance_attributes", "inputML", "outputML"]  # all datablock names

if get_names:

    for i, datablock_nickname in enumerate(datablocks):
        panel_title = datablock_nickname
        response = get_dataobject_names_from_block(session_id(), datablock_nickname)
        text_items = response["names"]
        if text_items:
            text_str = "\n".join(text_items)
            pt = sd.PointF(x + gap + i * gap, y + i * gap)
            rect = sd.RectangleF(0, 0, w, h)
            make_Panel(panel_title, text_str, pt, rect)
        else:
            errors += response["msg"] + "\n"

if errors:
    ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, str(errors))



# flake8: noqa
from aixd_ara.wrappers import WrapperShallowDataObject

if not dim or dim < 1:
    dim = 1


dobj = {"datatype": "DataBool", "name": name, "dim": dim}


dataobject = WrapperShallowDataObject(dobj)



# flake8: noqa

from scriptcontext import sticky as st

from aixd_ara.gh_ui import get_dataobject_types
from aixd_ara.gh_ui import request_designs
from aixd_ara.gh_ui_helper import component_id
from aixd_ara.gh_ui_helper import instantiate_sample
from aixd_ara.gh_ui_helper import sample_summary as sample_summary_f
from aixd_ara.gh_ui_helper import session_id
from aixd_ara.gh_ui_helper import reformat_request

cid = component_id(session_id(), ghenv.Component, "Generator")
item = component_id(session_id(), ghenv.Component, "GeneratorPickItem")

"""
requested_values: a multiline string with variable_name:values tuples.
"""

if not n_designs or n_designs < 1:
    n_designs = 1

# -------------------------------------------------------------------------------


if clear and cid in st.keys():
    del st[cid]
    st[item] = None
    ghenv.Component.Message = "no samples"

if generate and requested_values:

    variable_types = get_dataobject_types(session_id())["dataobject_types"]

    request_dict = reformat_request(requested_values, variable_types)
    st[item] = 0
    ghenv.Component.Message = "Running"
    st[cid] = request_designs(session_id(), request_dict, n_designs)
    ghenv.Component.Message = "Finished"

if pick_previous:
    st[item] -= 1
if pick_next:
    st[item] += 1

if cid in st.keys():
    samples = st[cid]["generated"]
    n = len(samples)
    i = st[item] % n

    ghenv.Component.Message = "sample {}/{}".format(i + 1, n)

    sample_dict = samples[i]
    ghdoc = ghenv.Component.OnPingDocument()
    instantiate_sample(ghdoc, sample_dict)  # sends design parameter values to the parametric model

    # --- output ---
    sample_summary = "Predicted/Generated:\n--------------------\n\n" + sample_summary_f(sample_dict)



# flake8: noqa

from scriptcontext import sticky as st

from aixd_ara.gh_ui import model_train
from aixd_ara.gh_ui import project_setup_info
from aixd_ara.gh_ui_helper import component_id
from aixd_ara.gh_ui_helper import session_id

cid = component_id(session_id(), ghenv.Component, "ModelTrain")


if not epochs or epochs < 1:
    epochs = 1

if wb:
    import webbrowser

    if run:
        projectname = project_setup_info(session_id())["project_name"]
        url = "https://wandb.ai/{}/{}".format(wb, projectname)
        webbrowser.open(url)
        # TODO: check login and add interface to login if needed


if run:
    st[cid] = model_train(session_id(), epochs, wb)


if cid in st.keys():
    best_ckpt = st[cid]["best_ckpt"]
    path = st[cid]["path"]
    ghenv.Component.Message = st[cid]["msg"]
else:
    ghenv.Component.Message = ""

# TODO:
# * return path or name of best checkpoint
# * make non-GUI-blocking
# * display progress



from aixd_ara.scripts import dataset_generator

path = dataset_generator.__file__

txt = "INSTRUCTIONS\n------------\n"
txt += "How to generate random samples for a dataset, "
txt += "by harnessing the parametric model defined in this Grasshopper file?\n\n"
txt += "Prepare the inputs and outputs to/from your parametric model as shown in the online documentation>tutorial at "
txt += "https://gramaziokohler.github.io/aixd_ara/latest/tutorial.html\n\n"
txt += "Close all other Grasshopper files in this Rhino app.\n\n"
txt += "Locate the 'dataset_generator.py' file in:\n\n"
txt += "{}\n\n".format(path)
txt += "In Rhino, open the Rhino Python Editor, open 'dataset_generator.py' and run it.\n"
txt += "(Follow the prompts in the Rhino command line)\n\n"

txt += "\n------------\n\n"
txt += "Optionally, add a button to a Rhino toolbar to access the dataset generator more easily:\n"
txt += "> right-click on a toolbar > New Button... \n"
txt += "> in the pop-up window, in Command section, paste:\n\n"
txt += "'! _-RunPythonScript {}'".format(path)

instructions = txt



# flake8: noqa
from scriptcontext import sticky as st

from aixd_ara.gh_ui import model_setup
from aixd_ara.gh_ui_helper import component_id
from aixd_ara.gh_ui_helper import session_id

cid = component_id(session_id(), ghenv.Component, "ModelSetup")

settings = {
    "model_type": model_type,
    "inputML": features,
    "outputML": targets,
    "latent_dim": latent_dim,
    "hidden_layers": hidden_layers,
    "batch_size": batch_size,
}


# TODO: make default settings a bit more smart
default_settings = {
    "model_type": "CAE",
    "inputML": "design_parameters",
    "outputML": "performance_attributes",
    "latent_dim": 8,
    "hidden_layers": [512, 256, 128, 64],
    "batch_size": 16,
}

for k in default_settings.keys():
    if (k not in settings) or (k in settings and settings[k] == None) or (k in settings and settings[k] == []):
        settings[k] = default_settings[k]
        print(k, default_settings[k])


if set:
    st[cid] = model_setup(session_id(), settings)


if cid in st.keys():
    quick_summary = st[cid]["quick_summary"]["summary"]
    model_dims = st[cid]["model_dims"]["summary"]
    ghenv.Component.Message = st[cid]["msg"]



# flake8: noqa
from scriptcontext import sticky as st

from aixd_ara.gh_ui import nn_summary
from aixd_ara.gh_ui_helper import component_id
from aixd_ara.gh_ui_helper import session_id

cid = component_id(session_id(), ghenv.Component, "ModelSummary")


if not max_depth:
    max_depth = 1

if get:
    st[cid] = nn_summary(session_id(), max_depth)

if cid in st.keys():
    summary = st[cid]["summary"]



# flake8: noqa
from scriptcontext import sticky as st

from aixd_ara.gh_ui import create_dataset_object
from aixd_ara.gh_ui_helper import component_id
from aixd_ara.gh_ui_helper import session_id

cid = component_id(session_id(), ghenv.Component, "DatasetCreate")


design_parameters = [x.data for x in design_parameters]
performance_attributes = [x.data for x in performance_attributes]


if create:
    st[cid] = create_dataset_object(session_id(), design_parameters, performance_attributes)

if cid in st.keys():
    msg = st[cid]["msg"]



# flake8: noqa
from aixd_ara.wrappers import WrapperShallowDataObject

if not dim or dim < 1:
    dim = 1

if not options:
    options = None
else:
    options = [str(x) for x in options]


dobj = {"datatype": "DataCategorical", "name": name, "dim": dim, "domain": options}

dataobject = WrapperShallowDataObject(dobj)



# flake8: noqa
from scriptcontext import sticky as st

from aixd_ara.gh_ui import model_input_output_dimensions
from aixd_ara.gh_ui_helper import component_id
from aixd_ara.gh_ui_helper import session_id

cid = component_id(session_id(), ghenv.Component, "ModelDims")

if get:
    st[cid] = model_input_output_dimensions(session_id())

if cid in st.keys():
    summary = st[cid]["summary"]



__author__ = ["Aleksandra Apolinarska"]
__copyright__ = "Copyright (c) 2023-2024 ETH Zurich, Gramazio Kohler Research"
__license__ = "MIT License"
__email__ = "apolinarska@arch.ethz.ch"
__version__ = "0.10.4"



# flake8: noqa
from scriptcontext import sticky as st
from aixd_ara.gh_ui import plot_contours_request
from aixd_ara.gh_ui import get_dataobject_types
from aixd_ara.gh_ui_helper import session_id
from aixd_ara.gh_ui_helper import component_id
from aixd_ara.gh_ui_helper import convert_str_to_bitmap
from aixd_ara.gh_ui_helper import reformat_request


cid = component_id(session_id, ghenv.Component, "PlotContoursRequest")


if plot:
    st[cid] = plot_contours_request(session_id(), "interactive") # will launch the plotly fig in browser

if cid in st.keys():
    print st[cid]


# flake8: noqa
from scriptcontext import sticky as st

from aixd_ara.gh_ui import get_one_sample
from aixd_ara.gh_ui_helper import component_id
from aixd_ara.gh_ui_helper import instantiate_sample, sample_summary
from aixd_ara.gh_ui_helper import session_id

cid = component_id(session_id(), ghenv.Component, "DatasetOneSample")


if not item:
    item = -1

if get:
    st[cid] = get_one_sample(session_id(), item)

ghdoc = ghenv.Component.OnPingDocument()


if cid in st.keys():
    sample = st[cid]
    instantiate_sample(ghdoc, sample)

    sample_summary = sample_summary(sample)



import json


import Grasshopper
import urllib2
from System import Convert
from System import Guid
from System.Drawing import Bitmap
from System.Drawing import Size
from System.IO import MemoryStream

from aixd_ara.constants import DEFAULT_PORT

TYPES = {
    "Arc": Grasshopper.Kernel.Types.GH_Arc,
    "Boolean": Grasshopper.Kernel.Types.GH_Boolean,
    "Box": Grasshopper.Kernel.Types.GH_Box,
    "Brep": Grasshopper.Kernel.Types.GH_Brep,
    "Circle": Grasshopper.Kernel.Types.GH_Circle,
    "ComplexNumber": Grasshopper.Kernel.Types.GH_ComplexNumber,
    "Curve": Grasshopper.Kernel.Types.GH_Curve,
    "Guid": Grasshopper.Kernel.Types.GH_Guid,
    "Integer": Grasshopper.Kernel.Types.GH_Integer,
    "Interval": Grasshopper.Kernel.Types.GH_Interval,
    "Interval2D": Grasshopper.Kernel.Types.GH_Interval2D,
    "Line": Grasshopper.Kernel.Types.GH_Line,
    "Mesh": Grasshopper.Kernel.Types.GH_Mesh,
    "Number": Grasshopper.Kernel.Types.GH_Number,
    "Plane": Grasshopper.Kernel.Types.GH_Plane,
    "Point": Grasshopper.Kernel.Types.GH_Point,
    "Rectangle": Grasshopper.Kernel.Types.GH_Rectangle,
    "String": Grasshopper.Kernel.Types.GH_String,
    "SubD": Grasshopper.Kernel.Types.GH_SubD,
    "Surface": Grasshopper.Kernel.Types.GH_Surface,
    "Vector": Grasshopper.Kernel.Types.GH_Vector,
    "Text": Grasshopper.Kernel.Types.GH_String,
}


def http_post_request(action, data):
    """
    action: string corresponding to the http header
    data: dictionary of json-serializable data
    """
    headers = {
        "Content-Type": "application/octet-stream",
        "Accept": "application/octet-stream",
    }
    data = json.dumps(data)
    url = "http://127.0.0.1:{}/{}".format(DEFAULT_PORT, action)
    request = urllib2.Request(url, data, headers)
    response = urllib2.urlopen(request).read()
    response = json.loads(response)

    return response


def session_id():
    doc_key = Grasshopper.Instances.ActiveCanvas.Document.DocumentID.ToString()
    return doc_key


def component_id(session_id, component, name):
    return "{}_{}_{}".format(session_id, component.InstanceGuid, name)


def clear_sticky(ghenv, st):
    """
    Removes all items from the sticky dictionary use by Grasshopper components in the given document.
    Resets all components that used the sticky to hold data.

    Parameters
    ----------
    ghenv: Grasshopper environment object `GhPython.Component.PythonEnvironment`
    st: sticky dictionary
    """

    ghdoc = ghenv.Component.OnPingDocument()
    ghdoc_id = ghdoc.DocumentID.ToString()

    keys = st.keys()

    # The keys we are looking for are strings of the form "{session_id}_{guid_str}_{ghcomponent_nickname}".
    # There might be other keys in the sticky dictionary, so we need to filter them out.
    for key in keys:

        session_id = None
        guid_str = None
        try:
            session_id = key.split("_")[0]
            guid_str = key.split("_")[1]
        except Exception:
            pass

        if not session_id or not guid_str:
            continue

        # The retrieved session_id and guid_str may either come from a different Grasshopper document,
        # or from some other process and be incorrect/meaningless.
        # In these cases, the following code will do nothing anyway.
        if session_id == ghdoc_id:
            reset_component(ghdoc, guid_str)
            st.pop(key)


def reset_component(ghdoc, guid_str):
    """
    adapted from:
    https://github.com/compas-dev/compas/blob/ea4b5b5191a350d24cbb479c6770daa68dbe53fd/src/compas_ghpython/timer.py#L8
    """

    guid = Guid(guid_str)
    ghcomp = ghdoc.FindComponent(guid)

    def callback(ghdoc):
        if ghdoc.SolutionState != Grasshopper.Kernel.GH_ProcessStep.Process:
            ghcomp.ExpireSolution(False)

    if not ghcomp:
        return
    delay = 1  # [ms]
    ghdoc.ScheduleSolution(delay, Grasshopper.Kernel.GH_Document.GH_ScheduleDelegate(callback))


def find_component_by_nickname(ghdoc, component_nickname):
    found = []
    all_objects = ghdoc.Objects
    # all_objects = ghenv.Component.OnPingDocument().Objects
    for obj in all_objects:
        if obj.Attributes.PathName == component_nickname:
            # if obj.NickName == component_nickname:
            found.append(obj)

    if not found:
        print("No ghcomponent found with a nickname {}.".format(component_nickname))
        return
    if len(found) > 1:
        print("{} ghcomponents found with the nickname {} - will return None.".format(len(found), component_nickname))
        return
    return found[0]


# set & get values methods (rhinopythonscript style)


def ghparam_set_values(component, values, expire=True):
    """
    Data type of values must match the type of the component.
    See TYPES list.
    """
    ghtype = TYPES[component.TypeName]

    component.Script_ClearPersistentData()
    if not isinstance(values, list):
        values = [values]
    for v in values:
        component.PersistentData.Append(ghtype(v))
    component.ExpireSolution(expire)


def ghparam_get_values(component, compute=False):
    if compute:
        component.CollectData()
        component.ComputeData()

    if not component.VolatileData:
        return None

    return [x.Value for x in component.VolatileData[0]]


def sample_summary(sample_dict):
    txt = ""
    txt += "Design Parameters:\n\n"
    for name, values in sample_dict["design_parameters"].items():
        txt += "{}:  {}\n".format(name, values)
    txt += "\n"
    txt += "Performance Attributes:\n\n"
    for name, values in sample_dict["performance_attributes"].items():
        txt += "{}:  {}\n".format(name, values)
    return txt


def instantiate_sample(ghdoc, sample_dict):
    """
    Apply design parameter values to the parametric model.
    """

    for dp_name, dp_vals in sample_dict["design_parameters"].items():
        component_name = "DP_{}".format(dp_name)
        component = find_component_by_nickname(ghdoc, component_name)

        if not dp_vals:
            print("No values for {}!".format(dp_name))
        else:
            ghparam_set_values(component, dp_vals)


def convert_str_to_bitmap(base64_imgstr, scale=1.0):
    """Get image from string and rescale."""

    b64_bytearray = Convert.FromBase64String(base64_imgstr)
    stream = MemoryStream(b64_bytearray)
    bitmap = Bitmap(stream)

    if not scale:
        scale = 1.0
    size = Size(bitmap.Width * scale, bitmap.Height * scale)
    bitmap = Bitmap(bitmap, size)
    return bitmap


def recast_type(value, typename):
    if typename == "real":
        return float(value)
    if typename == "integer":
        return int(value)
    if typename == "categorical":
        return str(value)
    if typename == "bool":
        return bool(value)


def reformat_request(request_string, variable_types):
    """
    Reformats the request string into a dictionary with the correct types.
    variable_types: dictionary with variable names as keys and types ('real', 'int', 'categorical', 'bool') as values,
    used to restore the data type.
    """
    request_dict = {}

    for rv in request_string:
        rv = rv.strip()

        # split into name:value(s)
        k, v = rv.split(":")

        if k not in variable_types.keys():
            raise ValueError(
                "'{0}' is not a valid variable name. There is not variable with this name in the dataset.".format(k)
            )

        # check if a list or a single value
        if v[0] == "[" and v[-1] == "]":
            v = v[1:-1]
            v = v.split(",")
            v = [recast_type(vi, variable_types[k]) for vi in v]
        else:
            v = recast_type(v, variable_types[k])
        request_dict[k] = v
    return request_dict



from compas.data import Data


class WrapperShallowDataObject(object):
    """
    Generic wrapper to store the shallow data object definitions as dictionaries.
    """

    def __init__(self, data):
        self.data = data

    def __repr__(self):
        return "{}: '{}' dim={} domain={}".format(
            self.data["datatype"], self.data["name"], self.data["dim"], self.data["domain"]
        )

    def __str__(self):
        return self.__repr__()


class WrapperGH(object):
    """
    Basic wrapper intended to help passing data around in Grasshopper.
    For example, it is useful when you want to pass a dictionary from one GH component to another.
    Without it, default behaviour of Grasshopper will pass only the dictionary keys as a list of strings.
    """

    def __init__(self, data):
        self.data = data


class WrapperSample(Data):
    """
    Simple wrapper class to store data of a single design sample.
    Intended to collect samples values for dataobjects.
    Each dictionary should have the following structure and contents:
    * keys : correspond to the data object names
    * values : List of values. For dataobject with dim=1, this is a 1-item list.
    """

    def __init__(self, dp={}, pa={}, dataobjects={}):
        self.design_parameters = dp
        self.performance_attributes = pa
        self.dataobjects = (
            dataobjects  # intended for an unsorted collection of dobj which is neither the collection in dp or pa
        )

    def __repr__(self):
        return "Sample Wrapper dp={} pa={} other={}".format(
            self.design_parameters, self.performance_attributes, self.dataobjects
        )



"""
Run this file in Rhino Python Editor or add it to your toolbar as a button using:
    > right-click on a toolbar > New Button... > in the pop-up window, in Command section, add:
    "! _-RunPythonScript "<path-to-this-file>/dataset_generator.py"
The Grasshopper file must be the first active file.
(Close all other Grasshopper files to avoid confusion.)
"""

import math
import os

import rhinoscriptsyntax as rs

from aixd_ara.gh_ui_helper import find_component_by_nickname
from aixd_ara.gh_ui_helper import ghparam_get_values
from aixd_ara.gh_ui_helper import ghparam_set_values
from aixd_ara.gh_ui_helper import http_post_request


# --- GRASSHOPPER INTERFACE ----------------------------------------------------------------------------

try:
    import Rhino

    Grasshopper = Rhino.RhinoApp.GetPlugInObject("Grasshopper")
    import Grasshopper
except:  # noqa: E722
    # because the first try sometimes doesn't work after "Reset Script Engine"
    import clr

    clr.AddReference("Grasshopper")
    import Grasshopper

docServer = Grasshopper.GH_InstanceServer.DocumentServer
ghdoc = docServer[0]  # first opened document
session_id = ghdoc.DocumentID.ToString()


# --- APP INTERFACE ----------------------------------------------------------------------------


def generate_dp_samples(n_samples):
    return http_post_request("generate_dp_samples", {"session_id": session_id, "n_samples": n_samples})


def calculate_pa_samples(ghdoc, dp_samples):
    pa_names = get_pa_names()
    pa_samples = analysis_callback(ghdoc, dp_samples, pa_names)
    return pa_samples


def load_dps():
    return http_post_request("getdata_design_parameters", {"session_id": session_id})


def get_pa_names():
    d = http_post_request("datablocks_dataobjects", {"session_id": session_id})
    return d["performance_attributes"]


def analysis_callback(ghdoc, dp_samples, pa_names):

    pa_samples = []

    # pass design parameters (sample by sample) to Grasshopper model and read the performance attributes
    for sample in dp_samples:
        for dp_name, dp_vals in sample.items():
            component_name = "DP_{}".format(dp_name)
            component = find_component_by_nickname(ghdoc, component_name)
            ghparam_set_values(component, dp_vals, expire=False)

        pa_dict = {k: [] for k in pa_names}
        for pa_name in pa_names:

            component_name = "PA_{}".format(pa_name)
            component = find_component_by_nickname(ghdoc, component_name)
            pa_vals = ghparam_get_values(component, compute=True)
            if isinstance(pa_vals, list):
                if len(pa_vals) == 1:
                    pa_vals = pa_vals[0]  # unpack from list
            pa_dict[pa_name] = pa_vals

        pa_samples.append(pa_dict)

    return pa_samples


def add_samples_to_dataset(samples, samples_per_file=None):
    return http_post_request(
        "save_samples", {"session_id": session_id, "samples": samples, "samples_per_file": samples_per_file}
    )


def combine_dp_pa(dp_samples, pa_samples):
    z = zip(dp_samples, pa_samples)  # one list with two dicts per entry
    samples = []
    for dp, pa in z:
        d = {}
        d.update(dp)
        d.update(pa)
        samples.append(d)
    return samples


# --- MAIN ----------------------------------------------------------------------------


def run(session_id, num_samples, num_samples_per_batch):

    num_batches = int(math.ceil(num_samples / num_samples_per_batch))

    # override the number of samples to be a multiple of the number of samples per batch
    num_samples = num_batches * num_samples_per_batch

    pr = http_post_request("project_setup_info", {"session_id": session_id})
    project_root = pr["project_root"]
    project_name = pr["project_name"]
    target_path = os.path.join(project_root, project_name)

    print(
        "\t ({} samples in {} batches will be generated and saved in {})".format(num_samples, num_batches, target_path)
    )

    for _ in range(num_batches):
        dp_samples = generate_dp_samples(num_samples_per_batch)
        pa_samples = calculate_pa_samples(ghdoc, dp_samples)
        samples = combine_dp_pa(dp_samples, pa_samples)
        add_samples_to_dataset(samples, num_samples_per_batch)

    print("\t successfully generated all {} samples in {} batch files".format(num_samples, num_batches))


# --- INPUT/USER INTERFACE ----------------------------------------------------------------------------


def get_user_input():

    num_samples = rs.GetInteger("Number of samples to generate: ", number=1000, minimum=1, maximum=None)
    num_samples_per_batch = rs.GetInteger(
        "Number of samples per batch file: ", number=num_samples / 10, minimum=1, maximum=num_samples
    )
    return num_samples, num_samples_per_batch


if __name__ == "__main__":

    num_samples, num_samples_per_batch = get_user_input()
    run(session_id, num_samples, num_samples_per_batch)



# flake8: noqa
from scriptcontext import sticky as st

from aixd_ara.gh_ui import plot_correlations
from aixd_ara.gh_ui_helper import component_id
from aixd_ara.gh_ui_helper import convert_str_to_bitmap
from aixd_ara.gh_ui_helper import session_id

cid = component_id(session_id(), ghenv.Component, "PlotCorrelations")


if plot:
    st[cid] = plot_correlations(session_id(), variables, "interactive")  # will launch the plotly fig in browser

if cid in st.keys():
    print(st[cid])


# flake8: noqa

from scriptcontext import sticky as st
from aixd_ara.gh_ui import reset_session
from aixd_ara.gh_ui_helper import session_id, clear_sticky

if reset:
    reset_session(session_id())
    clear_sticky(ghenv, st)



# flake8: noqa
from scriptcontext import sticky as st

from aixd_ara.gh_ui import model_load
from aixd_ara.gh_ui_helper import component_id
from aixd_ara.gh_ui_helper import session_id

cid = component_id(session_id(), ghenv.Component, "ModelLoad")

if not model_type:
    model_type = "CAE"
if not checkpoint_path:
    checkpoint_path = ""
if not checkpoint_name:
    checkpoint_name = "last"

if load:
    st[cid] = model_load(session_id(), model_type, checkpoint_name, checkpoint_path)

if cid in st.keys():
    msg = st[cid]["msg"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import urllib2

import compas
import compas._os
from ghpythonlib.componentbase import executingcomponent as component
from scriptcontext import sticky as st

from aixd_ara.constants import DEFAULT_PORT

try:
    from subprocess import PIPE
    from subprocess import Popen
except ImportError:
    from System.Diagnostics import Process

import os

import aixd_ara as ag


class ApiRunner(object):
    def __init__(
        self,
        package=None,
        python=None,
        url="http://127.0.0.1",
        port=DEFAULT_PORT,
        service=None,
        max_conn_attempts=100,
        autoreload=True,
        show_window=False,
        path=None,
        working_directory=None,
    ):
        self.python = compas._os.select_python(python)
        self.service = "aixd_ara.api"
        self.working_directory = working_directory
        self.show_window = show_window
        self.port = port

        self.start_process()

    def start_process(self):
        env = compas._os.prepare_environment()

        try:
            Popen
        except NameError:
            self._process = Process()
            for name in env:
                if self._process.StartInfo.EnvironmentVariables.ContainsKey(name):
                    self._process.StartInfo.EnvironmentVariables[name] = env[name]
                else:
                    self._process.StartInfo.EnvironmentVariables.Add(name, env[name])
            self._process.StartInfo.UseShellExecute = True
            self._process.StartInfo.RedirectStandardOutput = not self.show_window
            self._process.StartInfo.RedirectStandardError = not self.show_window
            self._process.StartInfo.FileName = self.python
            self._process.StartInfo.Arguments = "-m {0} {1}".format(self.service, self.port)
            self._process.Start()
        else:
            args = [self.python, "-m", self.service, str(self.port)]
            kwargs = dict(env=env)

            if self.show_window:
                if compas.is_mono():
                    python_command = " ".join(args)
                    apple_script = """
                    tell application "Terminal"
                        do script "{}"
                        activate
                    end tell
                    """.format(
                        python_command
                    )
                    args = ["osascript", "-e", apple_script]
            else:
                kwargs["stdout"] = PIPE
                kwargs["stderr"] = PIPE

            if self.working_directory:
                kwargs["cwd"] = self.working_directory
            self._process = Popen(args, **kwargs)

    def stop_process(self):
        """Attempts to terminate the python process hosting the proxy server.

        The process reference might not be present, e.g. in the case
        of reusing an existing connection. In that case, this is a no-op.
        """
        print("Stopping the server process.")

        if not self._process:
            return

        try:
            self._process.terminate()
        except Exception:
            pass
        try:
            self._process.kill()
        except Exception:
            pass


api_path = os.path.dirname(ag.__file__)

print(api_path)


def create_api_runner_id(component):
    name = "api-runner"
    return "{}_{}".format(name, component.InstanceGuid)


class ApiRunnerComponent(component):

    def RunScript(self, start, stop, show_window=True):
        key = create_api_runner_id(self)
        port = DEFAULT_PORT

        if show_window:
            python = "python"
        else:
            python = "pythonw"

        if start:
            self.stop(key)
            runner = ApiRunner(python=python, port=port, show_window=show_window, working_directory=api_path)
            st[key] = runner

        if stop:
            self.stop(key)

    def stop(self, key):
        # Attempt gracefully stopping the server
        try:
            urllib2.urlopen("http://127.0.0.1:8765/shutdown", timeout=1)
        except Exception:
            pass

        # stop previous API runner if any
        previous_api_runner = st.get(key)
        if previous_api_runner:
            try:
                previous_api_runner.stop_process()
            except Exception as e:
                print("Unable to stop previous process")
                print(e)
            st.pop(key)
        else:
            print("No previous process runner found")



from typing import Dict
from typing import List

from aixd.data.data_objects import DataBool
from aixd.data.data_objects import DataCategorical
from aixd.data.data_objects import DataInt
from aixd.data.data_objects import DataReal
from aixd.data.domain import Interval
from aixd.data.domain import Options

"""
Methods to convert between shallow versions of data objects and the proper DataObject classes.

A shallow version is a dictionary of a form:
dict = {'datatype': str,
        'name': str,
        'dim' : int,
        'domain' : list of
            [True, False] (for DataBool), or
            two values (float for DataReal), or
            n values (int for DataInt or any for DataCategorical), or
            None
    }

"""


def DataInt_from_shallow(shallow_dobj):
    return DataInt(
        name=shallow_dobj["name"],
        dim=shallow_dobj["dim"],
        domain=Interval(*shallow_dobj["domain"]) if shallow_dobj["domain"] else None,
    )


def DataReal_from_shallow(shallow_dobj):
    return DataReal(
        name=shallow_dobj["name"],
        dim=shallow_dobj["dim"],
        domain=Interval(*shallow_dobj["domain"]) if shallow_dobj["domain"] else None,
    )


def DataCategorical_from_shallow(shallow_dobj):
    return DataCategorical(
        name=shallow_dobj["name"],
        dim=shallow_dobj["dim"],
        domain=Options(shallow_dobj["domain"]) if shallow_dobj["domain"] else None,
    )


def DataBool_from_shallow(shallow_dobj):
    return DataBool(name=shallow_dobj["name"], dim=shallow_dobj["dim"])


def dataobjects_from_shallow(shallow_dobjs: List[Dict]):  # -> List[DataObject]:
    """
    Converts a list of shallo data objects to the proper instances of respective DataObject child classes.
    """
    dobj_list = []
    for sh_dobj in shallow_dobjs:
        if sh_dobj["datatype"] == "DataBool":
            dobj_list.append(DataBool_from_shallow(sh_dobj))
        elif sh_dobj["datatype"] == "DataCategorical":
            dobj_list.append(DataCategorical_from_shallow(sh_dobj))
        elif sh_dobj["datatype"] == "DataInt":
            dobj_list.append(DataInt_from_shallow(sh_dobj))
        elif sh_dobj["datatype"] == "DataReal":
            dobj_list.append(DataReal_from_shallow(sh_dobj))
        else:
            raise ValueError(
                f"Unknow data type {sh_dobj['datatype']} in a shallow data object."
                + "Allowed values (string) are: 'DataBool', 'DataCategorical', 'DataInt', 'DataReal'."
            )

    return dobj_list



# flake8: noqa
import os

from scriptcontext import sticky as st

from aixd_ara.gh_ui import project_setup
from aixd_ara.gh_ui_helper import clear_sticky
from aixd_ara.gh_ui_helper import component_id
from aixd_ara.gh_ui_helper import session_id

if not project_root:
    ghdoc = ghenv.Component.OnPingDocument().FilePath
    project_root = os.path.dirname(ghdoc)

if project_name:
    cid = component_id(session_id(), ghenv.Component, "ProjectSetup")

    if set:
        complete_path = os.path.join(project_root, project_name)
        if not os.path.exists(complete_path):
            print("The path {} does not exist and will be now created.".format(complete_path))
            os.mkdir(complete_path)

        clear_sticky(ghenv, st)

        st[cid] = project_setup(session_id(), project_root, project_name)

    if cid in st.keys():
        msg = st[cid]["msg"]
        path = st[cid]["path"]



"""
This module contains methods that are intended to run in cpython in a server app.
Do not call it from Rhino/Grasshopper (imports will fail in IronPython).
"""

import base64
import os
import random

import pytorch_lightning as pl
import torch
from aixd.data.data_blocks import DesignParameters
from aixd.data.data_blocks import PerformanceAttributes
from aixd.data.data_objects import DataInt, DataBool, DataCategorical, DataReal
from aixd.data.dataset import Dataset
from aixd.data.utils_data import reformat_dataframe_to_dataframeflat
from aixd.data.utils_data import reformat_dataframeflat_to_dict
from aixd.data.utils_data import reformat_dict_to_dataframe
from aixd.data.utils_data import reformat_dict_to_dictlist
from aixd.data.utils_data import reformat_dictlist_to_dict
from aixd.mlmodel.architecture.cond_ae_model import CondAEModel
from aixd.mlmodel.architecture.cond_vae_model import CondVAEModel
from aixd.mlmodel.data.data_loader import DataModule
from aixd.mlmodel.generation.generator import Generator
from aixd.utils.utils import flatten_dict
from aixd.visualisation.plotter import Plotter

from aixd_ara.shallow_objects import dataobjects_from_shallow


class SessionController(object):
    instances = {}

    def __init__(self):
        self.project_root = None
        self.project_name = None
        self.dataset = None
        self.model = None
        self.datamodule = None
        self.samples_per_file = None
        self.model_is_trained = False
        self.requested_designs = None, None, None

    def reset(self):
        self.project_root = None
        self.project_name = None
        self.dataset = None
        self.model = None
        self.datamodule = None
        self.samples_per_file = None
        self.model_is_trained = False
        self.requested_designs = None, None, None

    @property
    def dataset_path(self):
        if not self.project_root or not self.project_name:
            return None
        return os.path.join(self.project_root, self.project_name)

    @classmethod
    def create(cls, session_id):
        if session_id not in cls.instances:
            cls.instances[session_id] = cls()
        return cls.instances[session_id]

    def project_setup(self, project_root, project_name):
        if not os.path.exists(project_root):
            return {"msg": "Project path {project_root} does not exist!"}
        self.project_root = project_root
        self.project_name = project_name
        return {
            "msg": f"Project has been set up in: {os.path.join(self.project_root, self.project_name)}",
            "path": os.path.join(self.project_root, self.project_name),
        }

    def project_setup_info(self):
        return {"project_root": self.project_root, "project_name": self.project_name}

    def create_dataset_object(self, design_parameters, performance_attributes):
        """
        Creates a dataset object based on given definitions of dataobjects in shallow formatting.
        """

        if not self.project_root or not self.project_name:
            raise ValueError("You need to first set the project root path and the dataset name.")

        dp = DesignParameters(name="DP", dobj_list=dataobjects_from_shallow(design_parameters))
        pa = PerformanceAttributes(name="PA", dobj_list=dataobjects_from_shallow(performance_attributes))

        dataset = Dataset(name=self.project_name, design_par=dp, perf_attributes=pa, root_path=self.project_root)
        dataset.save_dataset_obj()

        # TODO: overrides an already assigned dataset - what about the datafiles if they exist?
        self.dataset = dataset

        return {"msg": "Dataset object has been created."}

    def generate_dp_samples(self, n_samples):
        if not self.dataset:
            raise ValueError("Dataset is not defined. Load or create a Dataset object first.")

        samples_dictlist = self.dataset.get_samples(n_samples=n_samples, format_out="dict_list")
        return samples_dictlist

    def save_samples(self, samples, samples_per_file):
        """
        Adds samples to dataset, saves them to files and saves the dataset object.
        Samples can be in any of the formats used in the project.
        """
        if not self.dataset:
            raise ValueError("Dataset is not defined. Load or create a Dataset object first.")
        if not samples_per_file:
            samples_per_file = None

        self.dataset.write_data_dp_pa(data_combined=samples, samples_perfile=samples_per_file)
        return True

    def getdata_design_parameters(self):
        if not self.dataset:
            raise ValueError("Dataset is not defined. Load or create a Dataset object first.")

        self.dataset.load()
        data = self.dataset.design_par.data
        dp_dict = reformat_dataframeflat_to_dict(data, self.dataset.design_par.dobj_list + [DataInt(name="uid", dim=1)])
        dp_dictlist = reformat_dict_to_dictlist(dp_dict)
        return dp_dictlist

    def datablocks_dataobjects(self):
        if not self.dataset:
            raise ValueError("Dataset is not defined. Load or create a Dataset object first.")
        blocks = {}
        blocks["design_parameters"] = self.dataset.design_par.names_list
        blocks["performance_attributes"] = self.dataset.perf_attributes.names_list
        return blocks

    @property
    def design_parameters_names(self):
        if not self.dataset:
            raise ValueError("Dataset is not defined. Load or create a Dataset object first.")
        return self.dataset.design_par.names_list

    def load_dataset(self):
        error = ""
        if not self.project_root or not self.project_name:
            error = "You need to first set the project root path and the dataset name."
            raise ValueError(error)
        try:
            dataset = Dataset(root_path=self.project_root, name=self.project_name, overwrite=False)
            dataset.load_dataset_obj()
            dataset.load()
            dataset.update_obj_domains(flag_only_perfatt=True)
        except:  # noqa: E722
            dataset = None
            error = "Loading dataset failed."
            raise ValueError(error)

        if not dataset:
            return {"msg": error}

        self.dataset = dataset

        id_to_open = self.dataset.data_gen_dp["fileid_vector"]
        report = "Loaded a total of {} samples from {} files".format(len(self.dataset.design_par.data), len(id_to_open))

        return {"msg": report}

    def import_data_from_dict(self, datadict, samples_per_file=None):
        """
        Imports data created elsewhere (e.g. performance attributes calculated in Grasshopper) and
        formated as a dictionary, to the dataset and saves to files.
        datadict: dictionary containing keys equal to object names and values are lists of data values.
                  format: list of n dictionaries datadict[nth_sample][object_name_as_key][ith_dimension]
                  TODO: must also contain an 'uid' key?
        """
        if not self.dataset:
            raise ValueError("Dataset is not loaded.")
        if not samples_per_file:
            samples_per_file = self.samples_per_file
        if not samples_per_file:
            raise ValueError(
                "Argument 'samples per file' is not specified (neither in the project nor given as argument here)."
            )

        datadict = reformat_dictlist_to_dict(datadict)
        dataobjects = [d for d in self.dataset.data_objects if d.name in datadict.keys()]
        df = reformat_dict_to_dataframe(datadict)
        df = reformat_dataframe_to_dataframeflat(df, dataobjects)
        self.dataset.import_data_from_df(data=df, samples_perfile=samples_per_file)

        return True  # confirm it went well

    def dataset_summary(self):
        error = ""
        if not self.dataset:
            error = "Dataset is not loaded."
            raise ValueError(error)

        # flag_only_names = False
        txt = "-------------------------------------\n"
        txt += "Data blocks and elements in dataset\n\n"

        txt += "* Design parameters\n"
        for x in [
            f"{dobj.name} dim={str(dobj.dim)} domain: {dobj.domain}" for dobj in self.dataset.design_par.dobj_list
        ]:
            txt += f"    {x}\n"

        txt += "\n"
        txt += "* Performance attributes\n"
        for x in [
            f"{dobj.name} dim={str(dobj.dim)} domain: {dobj.domain}" for dobj in self.dataset.perf_attributes.dobj_list
        ]:
            txt += f"    {x}\n"

        txt += "-------------------------------------"
        return {"msg": error, "summary": txt}

    def get_dataobject_names_from_block(self, datablock_nickname):
        if datablock_nickname in ["design_parameters", "performance_attributes"] and not self.dataset:
            return {"msg": "Dataset is not loaded.", "names": []}
        if datablock_nickname in ["inputML", "outputML"] and not self.datamodule:
            return {"msg": "Model is not loaded.", "names": []}

        if datablock_nickname == "design_parameters":
            return {"msg": "", "names": self.dataset.design_par.names_list}
        if datablock_nickname == "performance_attributes":
            return {"msg": "", "names": self.dataset.perf_attributes.names_list}
        if datablock_nickname == "inputML":
            return {"msg": "", "names": self.datamodule.input_ml_dblock.names_list}
        if datablock_nickname == "outputML":
            return {"msg": "", "names": self.datamodule.output_ml_dblock.names_list}
        return {"msg": f"Wrong block nickname: {datablock_nickname}.", "names": []}

    def cast_to_python_type(self, dataobject_name, value):
        """
        Cast the values (usually coming from a dataframe) to the correct python type.
        Value can be a single value or a list.
        """
        if not self.dataset:
            raise ValueError("Dataset is not loaded.")

        dobj = self.dataset.get_data_objects_by_name([dataobject_name])[0]

        if not isinstance(value, list):
            value = [value]
            single = True
        else:
            single = False

        if isinstance(dobj, DataInt):
            castvalue = [int(v) for v in value]
        elif isinstance(dobj, DataReal):
            castvalue = [float(v) for v in value]
        elif isinstance(dobj, DataBool):
            castvalue = [bool(v) for v in value]
        elif isinstance(dobj, DataCategorical):
            castvalue = [str(v) for v in value]
        else:
            raise ValueError(f"Dataobject type not recognized: {dobj.type}")

        if single:
            castvalue = castvalue[0]
        return castvalue

    def get_dataobject_types(self):
        """
        Returns names of the data types of the dataobjects in the dataset.
        """
        if not self.dataset:
            error = "Dataset is not loaded."
            raise ValueError(error)

        all_dataobjects = self.dataset.data_objects
        # cannot use this because boolean are declared as categorical
        # dataobject_types = {d.name: d.type for d in all_dataobjects}

        dataobject_types = {}
        for d in all_dataobjects:
            if isinstance(d, DataInt):
                dataobject_types[d.name] = "integer"
            elif isinstance(d, DataReal):
                dataobject_types[d.name] = "real"
            elif isinstance(d, DataCategorical):
                dataobject_types[d.name] = "categorical"
            elif isinstance(d, DataBool):
                dataobject_types[d.name] = "boolean"
            else:
                dataobject_types[d.name] = "unsupported"

        return {"msg": "", "dataobject_types": dataobject_types}

    def get_design_parameters(self):
        # TODO: rename
        if not self.dataset:
            raise ValueError("Dataset is not loaded.")
        print(self.dataset.design_par)
        dp = self.dataset.design_par.names_list
        return dp

    def plot_distrib_attributes(self, dataobjects, output_type):
        """
        dataobjects: list of dataobject names
        """
        if not self.dataset:
            raise ValueError("Dataset is not loaded.")

        if not dataobjects:
            block = self.dataset.perf_attributes.name  # by default, we're interested in performance attributes here
        else:
            block = self.blocknames_from_dataobjects(dataobjects)
            if len(block) > 1:
                raise ValueError("Dataobjects are not from the same block.")
            if len(block) == 0:
                raise ValueError("Dataobjects are not from any block.")

        plotter = Plotter(self.dataset, output=None)
        fig = plotter.distrib_attributes(
            block=block[0], attributes=dataobjects, per_column=True, bottom_top=(0.1, 0.9), downsamp=1, sub_figs=True
        )
        return _fig_output(fig, output_type)

    def plot_correlations(self, dataobjects, output_type):
        """
        blocks and dataobjects: lists of names
        """
        if not self.dataset:
            raise ValueError("Dataset is not loaded.")

        if not dataobjects:
            dataobjects = [d.name for d in self.dataset.data_objects]

        blocks = self.blocknames_from_dataobjects(dataobjects)

        plotter = Plotter(self.dataset, output=None)
        fig = plotter.correlation(block=blocks, attributes=dataobjects)
        return _fig_output(fig, output_type)

    def plot_contours(self, dataobjects, output_type):
        """
        dataobjects: list of dataobject names
        """
        if not self.dataset:
            raise ValueError("Dataset is not loaded.")
        block = self.blocknames_from_dataobjects(dataobjects)[0]

        plotter = Plotter(self.dataset, output=None)
        fig = plotter.contours2d(block=block, attributes=dataobjects)
        return _fig_output(fig, output_type)

    def plot_contours_request(self, output_type):
        """
        request:
            dictionary where keys are the names of dataobjects (usually performance attributes),
            and values the requested target value(s).
        """
        if not self.dataset:
            raise ValueError("Dataset is not loaded.")
        if not self.model:
            raise ValueError("Model is not loaded.")
        if self.requested_designs == (None, None, None):
            raise ValueError("No designs have been requested yet.")

        plotter = Plotter(datamodule=self.datamodule, output=None)
        detailed_results = self.requested_designs[2]
        n_samples = len(self.requested_designs[1])

        fig = plotter.generation_scatter([detailed_results], n_samples=n_samples)
        return _fig_output(fig, output_type)

    def model_setup(self, model_type, inputML, outputML, latent_dim, layer_widths, batch_size):
        # TODO: set defaults here if missing?
        if not self.dataset:
            raise ValueError("Dataset is not loaded.")

        # TODO: move this check to the resp. datablocks so that
        # they recognize "design_parameters" and "performance_attributes" as argument?
        if inputML == ["design_parameters"]:
            inputML = self.dataset.design_par.names_list
        if outputML == ["design_parameters"]:
            outputML = self.dataset.design_par.names_list
        if inputML == ["performance_attributes"]:
            inputML = self.dataset.perf_attributes.names_list
        if outputML == ["performance_attributes"]:
            outputML = self.dataset.perf_attributes.names_list

        datamodule = DataModule.from_dataset(
            self.dataset, input_ml_names=inputML, output_ml_names=outputML, batch_size=batch_size
        )
        self.datamodule = datamodule

        save_dir = self.dataset_path

        if model_type == "CAE":
            model = CondAEModel.from_datamodule(
                datamodule, layer_widths=layer_widths, latent_dim=latent_dim, save_dir=save_dir
            )
        elif model_type == "CVAE":
            model = CondVAEModel.from_datamodule(
                datamodule, layer_widths=layer_widths, latent_dim=latent_dim, save_dir=save_dir
            )
        else:
            raise ValueError("Model type not recognized. Choose 'CAE' or 'CVAE'.")

        self.model = model
        self.model_is_trained = False

        quick_summary = self.model_summary(self.model, max_depth=2)
        model_dims = self.model_input_output_dimensions()
        return {"msg": "Model has been set up.", "quick_summary": quick_summary, "model_dims": model_dims}

    def model_train(self, epochs, wb):
        if not self.model:
            raise ValueError("Model is not set up. Try setting up a model first.")
        self.model_is_trained = False

        if not wb:
            log_wb = False
        else:
            log_wb = True

        self.model.fit(
            self.datamodule,
            name_run="",
            max_epochs=epochs,
            callbacks=[],
            accelerator="cpu",
            flag_wandb=log_wb,
            wandb_entity=wb,
        )
        self.model_is_trained = True
        # TODO: store the best model in controller instead?
        checkpoint_path = os.path.join(self.model.save_dir, self.model.CHECKPOINT_DIR)

        # TODO: add some callback so that we can have a progress preview in Grasshopper
        # TODO: still saving the checkpoints in strange locations!!! return path to best checkpoint
        # TODO: add retrieve and return the name/path of the best checkpoint

        return {"msg": "Training completed!", "path": checkpoint_path, "best_ckpt": None}

    def model_load(self, model_type, checkpoint_path, checkpoint_name):
        error = None
        if checkpoint_path not in [None, ""]:
            if not os.path.exists(checkpoint_path):
                error = f"The given checkpoint path does not exist: {checkpoint_path}"
                raise ValueError(error)
        else:
            # default to the project path
            checkpoint_path = os.path.join(self.dataset_path, "checkpoints")

        checkpoint_filepath = os.path.join(checkpoint_path, checkpoint_name + ".ckpt")
        if not os.path.exists(checkpoint_path):
            error = f"The given checkpoint path does not exist: {checkpoint_filepath}"
            raise ValueError(error)

        if model_type == "CAE":
            model = CondAEModel.load_model_from_checkpoint(checkpoint_filepath)
        elif model_type == "CVAE":
            model = CondVAEModel.load_model_from_checkpoint(checkpoint_filepath)
        else:
            raise ValueError("Model type not recognized. Choose 'CAE' or 'CVAE'.")

        self.model = model
        self.model_is_trained = True
        self.datamodule = self._datamodule_from_dataset()
        return {"msg": error or f"Model loaded from checkpoint: {checkpoint_filepath}"}

    def _datamodule_from_dataset(self):
        if not self.dataset:
            raise ValueError("Dataset is not loaded.")

        datamodule = DataModule.from_dataset(
            self.dataset,
            input_ml_names=self.model.datamodule_parameters["input_ml_dblock"].names_list,
            output_ml_names=self.model.datamodule_parameters["output_ml_dblock"].names_list,
            batch_size=self.model.datamodule_parameters["batch_size"],
        )
        return datamodule

    def get_one_sample(self, item):
        """
        Returns a single sample from the dataset as a dictionary.

        Parameters
        ----------
        item : int >=-1 or None
            Index of the sample. If None or -1, a random sample will be drawn.

        """
        if not self.dataset:
            raise ValueError("Dataset is not loaded.")

        if item is None or item < 0:
            n = len(self.dataset.design_par.data)
            item = random.randint(0, n)

        sample = {"design_parameters": {}, "performance_attributes": {}}

        dct = reformat_dataframeflat_to_dict(self.dataset.design_par.data, self.dataset.design_par.dobj_list)
        for key, values in dct.items():
            dct[key] = self.cast_to_python_type(key, values[item])
        sample["design_parameters"] = dct

        dct = reformat_dataframeflat_to_dict(self.dataset.perf_attributes.data, self.dataset.perf_attributes.dobj_list)
        for key, values in dct.items():
            dct[key] = self.cast_to_python_type(key, values[item])
        sample["performance_attributes"] = dct

        # for single-value entries, unpack them from a list [123] -> 123
        for x in sample.keys():
            for k, v in sample[x].items():
                if isinstance(v, list):
                    if len(v) == 1:
                        sample[x][k] = v[0]
        return sample

    def request_designs(self, request, n_samples=1):
        """
        Parameters
        ----------
        n: int
            number of designs to return
        request: [Dict]
            dictionary where keys are the names of dataobjects (usually performance attributes),
            and values are the requested target values to fulfil
            Example:
            request = {"attribute1": 123,"attribute2":[45.0,67.0] }

        Returns
        -------
        List[Dict]
            List containing generated samples. Each sample is represented by a dictionary containing
            dataobject names as keys and values (requested, generated or predicted)

        """
        if not self.dataset:
            raise ValueError("Dataset is not loaded.")
        if not self.model:
            raise ValueError("Model is not loaded.")

        gen = Generator(model=self.model, datamodule=self.datamodule, over_sample=100)
        new_designs, detailed_results = gen.generate(request=request, n_samples=n_samples, format_out="dict_list")
        self.requested_designs = request, new_designs, detailed_results
        # split the result into separate dictionaries for design parameters and performance attributes
        # assert len(new_designs) == n_samples
        samples = []
        for d in new_designs:
            s = {"design_parameters": {}, "performance_attributes": {}}
            for k, v in d.items():
                if k in self.dataset.design_par.names_list:
                    s["design_parameters"][k] = v
                if k in self.dataset.perf_attributes.names_list:
                    s["performance_attributes"][k] = v
            samples.append(s)

        return {"msg": "", "generated": samples}

    def model_summary(self, model=None, max_depth=-1):

        if not model:
            model = self.model
        if not model:
            raise ValueError("No NN model given or loaded.")

        model.example_input_array = (
            (
                torch.zeros(
                    1,
                    max([x_dobj.position_index + x_dobj.dim for x_dobj in model.input_ml_dblock.dobj_list_transf]),
                ),
                torch.zeros(
                    1,
                    max([y_dobj.position_index + y_dobj.dim for y_dobj in model.input_ml_dblock.dobj_list_transf]),
                ),
            ),
        )
        summary = str(pl.utilities.model_summary.ModelSummary(model, max_depth=max_depth))
        return {"summary": summary}

    def model_input_output_dimensions(self):
        if not self.datamodule:
            raise ValueError("DataModule is not loaded. Try loading a model first.")

        inputdim, outputdim, summary = self.datamodule.summary_input_output_dimensions()
        return {"msg": "", "summary": summary}

    def all_block_names(self):
        """
        Returns all block names in the dataset.
        """
        if not self.dataset:
            raise ValueError("Dataset is not loaded.")

        datablocks = flatten_dict(self.dataset.data_blocks)
        names = [db.name for db in datablocks]
        return names

    def blocknames_from_dataobjects(self, dataobjects):
        """
        Returns the name of the block that contains the given dataobjects (given by their name).
        """
        datablocks = flatten_dict(self.dataset.data_blocks)  # list of all blocks

        blocknames = []
        for block in datablocks:
            for dobjname in dataobjects:
                if dobjname in block.names_list:
                    blocknames.append(block.name)

        return list(set(blocknames))


# --------------------------------------------------------------
# helper methods
# --------------------------------------------------------------


def _fig_output(fig, output_type):
    if not fig:
        return {"msg": "Plot failed."}

    if output_type == "static":
        imgstr = _fig_to_str(fig)
        return {"msg": "Static plot has been created.", "imgstr": imgstr}

    elif output_type == "interactive":
        fig.show()
        return {"msg": "Interactive plot has been launched in a new browser window."}


def _fig_to_str(fig):
    """
    Convert a plotly figure graph to a string-encoded bytes.
    """
    img_bytes = base64.b64encode(fig.to_image())
    img_string = img_bytes.decode("utf-8")
    return img_string



# flake8: noqa
from scriptcontext import sticky as st

from aixd_ara.gh_ui import load_dataset
from aixd_ara.gh_ui_helper import component_id
from aixd_ara.gh_ui_helper import session_id

cid = component_id(session_id(), ghenv.Component, "DatasetLoad")

if load:
    st[cid] = load_dataset(session_id())

if cid in st.keys():
    msg = st[cid]["msg"]



from aixd_ara.gh_ui_helper import http_post_request


def reset_session(session_id):
    data = {"session_id": session_id}
    return http_post_request(action="reset_session", data=data)


def project_setup(session_id, project_root, project_name):
    data = {"session_id": session_id, "project_root": project_root, "project_name": project_name}
    return http_post_request(action="project_setup", data=data)


def project_setup_info(session_id):
    data = {"session_id": session_id}
    return http_post_request(action="project_setup_info", data=data)


def create_dataset_object(session_id, design_parameters, performance_attributes):
    data = {
        "session_id": session_id,
        "design_parameters": design_parameters,
        "performance_attributes": performance_attributes,
    }
    return http_post_request(action="create_dataset_object", data=data)


def load_dataset(session_id):
    data = {"session_id": session_id}
    return http_post_request(action="load_dataset", data=data)


def dataset_summary(session_id):
    data = {"session_id": session_id}
    return http_post_request(action="dataset_summary", data=data)


def get_dataobject_names_from_block(session_id, datablock_nickname):
    data = {"session_id": session_id, "datablock_nickname": datablock_nickname}
    return http_post_request(action="get_dataobject_names_from_block", data=data)


def get_dataobject_types(session_id):
    data = {"session_id": session_id}
    return http_post_request(action="get_dataobject_types", data=data)


def get_one_sample(session_id, i):
    data = {"session_id": session_id, "item": i}
    return http_post_request(action="get_one_sample", data=data)


def plot_distrib_attributes(session_id, attributes, output_type):
    data = {"session_id": session_id, "attributes": attributes, "output_type": output_type}
    return http_post_request(action="plot_distrib_attributes", data=data)


def plot_contours(session_id, attributes, output_type):
    data = {"session_id": session_id, "attributes": attributes, "output_type": output_type}
    return http_post_request(action="plot_contours", data=data)


def plot_contours_request(session_id, output_type):
    data = {"session_id": session_id, "output_type": output_type}
    return http_post_request(action="plot_contours_request", data=data)


def plot_correlations(session_id, attributes, output_type):
    data = {"session_id": session_id, "attributes": attributes, "output_type": output_type}
    return http_post_request(action="plot_correlations", data=data)


def model_setup(session_id, settings):
    data = {"session_id": session_id, "settings": settings}
    return http_post_request(action="model_setup", data=data)


def model_train(session_id, epochs, wb):
    data = {"session_id": session_id, "epochs": epochs, "wb": wb}
    return http_post_request(action="model_train", data=data)


def model_load(session_id, model_type, checkpoint_name, checkpoint_path):
    data = {
        "session_id": session_id,
        "model_type": model_type,
        "checkpoint_name": checkpoint_name,
        "checkpoint_path": checkpoint_path,
    }
    return http_post_request(action="model_load", data=data)


def nn_summary(session_id, max_depth):
    data = {"session_id": session_id, "max_depth": max_depth}
    return http_post_request(action="nn_summary", data=data)


def model_input_output_dimensions(session_id):
    data = {"session_id": session_id}
    return http_post_request(action="model_input_output_dimensions", data=data)


def request_designs(session_id, request, n_designs):
    data = {"session_id": session_id, "requested_values": request, "n_designs": n_designs}
    return http_post_request(action="request_designs", data=data)



import json
import os
import signal

from compas.data import DataEncoder
from flask import Flask
from flask import request

from aixd_ara.constants import DEFAULT_PORT
from aixd_ara.controller import SessionController

app = Flask(__name__)


@app.post("/reset_session")
def reset_session():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)
    sc.reset()
    result = {"msg": "Session reset"}
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/project_setup", methods=["POST"])
def project_setup():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    sc.reset()  # Reset the session
    project_root = data["project_root"]
    project_name = data["project_name"]
    result = sc.project_setup(project_root, project_name)
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/project_setup_info", methods=["POST"])
def project_setup_info():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    result = sc.project_setup_info()
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/create_dataset_object", methods=["POST"])
def create_dataset_object():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    design_parameters = data["design_parameters"]
    performance_attributes = data["performance_attributes"]
    result = sc.create_dataset_object(design_parameters, performance_attributes)
    response = json.dumps(result, cls=DataEncoder)
    return str(response)


@app.route("/generate_dp_samples", methods=["POST"])
def generate_dp_samples():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    n_samples = data["n_samples"]
    result = sc.generate_dp_samples(n_samples)
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/save_samples", methods=["POST"])
def save_samples():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    n_samples = data["samples"]
    samples_per_file = data["samples_per_file"]
    result = sc.save_samples(n_samples, samples_per_file)
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/load_dataset", methods=["POST"])
def load_dataset():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    result = sc.load_dataset()
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/dataset_summary", methods=["POST"])
def dataset_summary():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    result = sc.dataset_summary()
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/getdata_design_parameters", methods=["POST"])
def getdata_design_parameters():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    result = sc.getdata_design_parameters()
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/import_data_from_dict", methods=["POST"])
def import_data_from_dict():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    datadict = data["datadict"]
    result = sc.import_data_from_dict(datadict)
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/datablocks_dataobjects", methods=["POST"])
def datablocks_dataobjects():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    result = sc.datablocks_dataobjects()
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/get_dataobject_names_from_block", methods=["POST"])
def get_dataobject_names_from_block():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    datablock_nickname = data["datablock_nickname"]
    result = sc.get_dataobject_names_from_block(datablock_nickname)
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/get_dataobject_types", methods=["POST"])
def get_dataobject_types():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    result = sc.get_dataobject_types()
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/plot_distrib_attributes", methods=["POST"])
def plot_distrib_attributes():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    output_type = data["output_type"]
    attributes = data["attributes"]
    result = sc.plot_distrib_attributes(dataobjects=attributes, output_type=output_type)
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/plot_correlations", methods=["POST"])
def plot_correlations():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    output_type = data["output_type"]
    attributes = data["attributes"]
    result = sc.plot_correlations(dataobjects=attributes, output_type=output_type)
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/plot_contours", methods=["POST"])
def plot_contours():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    output_type = data["output_type"]
    attributes = data["attributes"]
    result = sc.plot_contours(dataobjects=attributes, output_type=output_type)
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/plot_contours_request", methods=["POST"])
def plot_contours_request():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)
    output_type = data["output_type"]

    result = sc.plot_contours_request(output_type=output_type)
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/design_parameters", methods=["GET"])
def get_design_parameters():
    data = request.args
    session_id = data["session_id"]

    sc = SessionController.create(session_id)
    result = sc.get_design_parameters()
    return result


@app.post("/model_setup")
def model_setup():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    settings = data["settings"]
    model_type = settings["model_type"]
    inputML = settings["inputML"]
    outputML = settings["outputML"]
    latent_dim = settings["latent_dim"]
    hidden_layers = settings["hidden_layers"]
    batch_size = settings["batch_size"]

    result = sc.model_setup(model_type, inputML, outputML, latent_dim, hidden_layers, batch_size)
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.post("/model_train")
def model_train():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    epochs = data["epochs"]
    wb = data["wb"]
    result = sc.model_train(epochs, wb)
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/model_load", methods=["POST"])
def model_load():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    model_type = data["model_type"]
    checkpoint_path = data["checkpoint_path"]
    checkpoint_name = data["checkpoint_name"]
    result = sc.model_load(model_type, checkpoint_path=checkpoint_path, checkpoint_name=checkpoint_name)
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/request_designs", methods=["POST"])
def request_designs():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    requested_values = data["requested_values"]
    n_designs = data["n_designs"]
    result = sc.request_designs(requested_values, n_designs)
    result = json.dumps(result, cls=DataEncoder)
    return result


@app.route("/get_one_sample", methods=["POST"])
def get_one_sample():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    item = data["item"]
    result = sc.get_one_sample(item)
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/nn_summary", methods=["POST"])
def nn_summary():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    max_depth = data["max_depth"]
    result = sc.model_summary(max_depth=int(max_depth))
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/model_input_output_dimensions", methods=["POST"])
def model_input_output_dimensions():
    data = request.data
    data = json.loads(data)
    session_id = data["session_id"]
    sc = SessionController.create(session_id)

    result = sc.model_input_output_dimensions()
    response = json.dumps(result, cls=DataEncoder)
    return response


@app.route("/shutdown", methods=["GET"])
def shutdown():
    os.kill(os.getpid(), signal.SIGINT)
    print()
    print()
    print()
    print("*********************************************************")
    print("THIS APP HAS TERMINATED AND THIS WINDOW CAN BE CLOSED NOW")
    print("*********************************************************")
    print()
    print()
    print()
    return "Server shuting down"


@app.route("/ping", methods=["GET"])
def ping():
    return "Pong!"


def ara_welcome(*args):
    print()
    print()
    print()
    print("****************************")
    print("HELLO! ARA SERVER IS RUNNING")
    print("****************************")
    print()
    print()
    print()


if __name__ == "__main__":
    import sys

    if len(sys.argv) == 1:
        port = DEFAULT_PORT
    else:
        try:
            port = int(sys.argv[1])
        except ValueError:
            print("Invalid port number: ", sys.argv[1], "Setting a default port number {}".format(DEFAULT_PORT))
            port = DEFAULT_PORT

    cli = sys.modules["flask.cli"]
    cli.show_server_banner = ara_welcome

    app.run(host="127.0.0.1", port=port, debug=False)



# flake8: noqa
from scriptcontext import sticky as st

from aixd_ara.gh_ui import dataset_summary
from aixd_ara.gh_ui_helper import component_id
from aixd_ara.gh_ui_helper import session_id

cid = component_id(session_id(), ghenv.Component, "DatasetSummary")

if get:
    st[cid] = dataset_summary(session_id())

if cid in st.keys():
    if st[cid]["msg"]:
        summary = st[cid]["msg"]  # error
    else:
        summary = st[cid]["summary"]



"""Reveals the folder in the file explorer."""
# flake8: noqa

import compas
import subprocess

if open:
    command = "open"
    if compas.is_windows():
        command = "explorer"

    subprocess.Popen([command, path])
